/* equal.c */

#include "mlvalues.h"
#include "mem_misc.h"
#include "memory.h"
#include "stat_alloc.h"
#include "fail.h"

/* Egalite structurelle d'arbres.
   Ne termine pas sur les structures cycliques. */

static int tree_equal_aux(v1, v2)
     value v1,v2;
{
  blocksize i;
  value * p1, * p2;

 again:
  if (v1 == v2) return 1;
  if (IS_INT(v1) || IS_INT(v2) || !IS_ALLOCATED(v1) || !IS_ALLOCATED(v2))
    return 0;
  if (HEADER(v1) != HEADER(v2)) return 0;
  switch(TAG(v1)) {
  case TAG_STRING:
    return (compare_strings(v1, v2) == MLINT(0));
  case TAG_FLOAT:
    return (CFLOAT(v1) == CFLOAT(v2));
  case TAG_BITVECT:
    for(i = SIZE(v1), p1 = POINTER(v1), p2 = POINTER(v2);
        i > 0;
        i--, p1++, p2++)
      if (*p1 != *p2) return 0;
    return 1;
  default:
    for(i = SIZE(v1), p1 = POINTER(v1), p2 = POINTER(v2);
        i > 1;
        i--, p1++, p2++)
      if (!tree_equal_aux(*p1, *p2)) return 0;
    v1 = *p1;
    v2 = *p2;                   /* Recursion terminale */
    goto again;
  }
}

value tree_equal(v1, v2) /* ML */
     value v1, v2;
{
  return atom(tree_equal_aux(v1,v2));
}

