/* main.c : all together, now ! */

#include <fcntl.h>
#include <setjmp.h>
#include "mlvalues.h"
#include "alloc.h"
#include "stacks.h"
#include "globals.h"
#include "mem_misc.h"
#include "exec.h"
#include "fail.h"
#ifdef DEBUG
#include <stdio.h>
#endif

header first_atoms[256];

static void init_atoms()
{
  int i;
  for(i = 0; i < 256; i++) first_atoms[i] = MAKE_HEADER(0, i);
}

static void read_header(fd, hdr)
     int fd;
     struct exec_header * hdr;
{
  unsigned char header_buffer[EXEC_COMMENT_SIZE + sizeof(struct exec_header)];
  unsigned char * p;
  int n;
  
  n = read(fd, header_buffer, EXEC_COMMENT_SIZE + sizeof(struct exec_header));
  p = header_buffer;
  while (*p != '\012') {
    p++;
    n--;
    if (n <= 0) goto truncated;
  }
  p++;
  n--;

#define READ_INT(dest) {                                                      \
  if (n < 4) goto truncated;                                                  \
  dest = (p[0] << 24) + (p[1] << 16) + (p[2] << 8) + p[3];                    \
  n -= 4;                                                                     \
  p += 4;                                                                     \
}
  READ_INT(hdr->code_size);
  READ_INT(hdr->data_size);
  READ_INT(hdr->symbol_size);
  READ_INT(hdr->debug_size);

#undef READ_INT

  lseek(fd, p - header_buffer, 0);
  return;

 truncated:
  fatal_error("Truncated bytecode file");
}

code start_code;

static void read_code(fd, size)
     int fd;
     unsigned long size;
{
  start_code = (code) stat_alloc(size);
  if (read(fd, start_code, size) != size)
    fatal_error("Truncated bytecode file");
}

static void read_global_data(fd, size)
     int fd;
     unsigned long size;
{
  value block;
  long block_size;
  long n;

  if (size < 12 || size & 3 != 0)
    goto truncated;
  size -= 4;                    /* Le dernier mot est un ptr vers le res. */
  block_size = (size >> 2) - 1; /* Taille du bloc en mots */
  block = alloc(block_size, TAG_BITVECT);
  if (read(fd, (char *) ADDRESS(block), size) != size)
    goto truncated;
  adjust_pointers(block, block_size+1);
  if (read(fd, (char *) &n, sizeof(n)) != sizeof(n))
    goto truncated;
  global_data = block - 4 + n;
  
  FIELD(global_data, GLOBAL_DATA) = global_data; /* Ne peut pas creer de
                                                    racines inverses */
  return;

 truncated:
  fatal_error("Truncated bytecode file");
}

char usage[] =
  "usage: runtime [-v] [-m total size] [-h heap size] [-g generation size] \n               [-s static size] [-a arg stack size] [-r ret stack size] <file>\n";


int main(argc, argv)
     int argc;
     char * argv[];
{
  struct exec_header hdr;
  int fd;
  int i;
  unsigned long n;
  unsigned long total_size, heap_size, generation_size, static_size;
  unsigned long arg_stack_size, ret_stack_size;
  jmp_buf raise_buf;
  extern jmp_buf * external_raise;

#if defined(DEBUG) && defined(UNIX_MEMORY)
  char stdout_buffer[BUFSIZ];
  setbuf(stdout, stdout_buffer); /* Pour le debug et les messages d'erreurs */
  setlinebuf(stdout);            /* uniquement */
#endif

  total_size = TOTAL_SIZE;
  heap_size = HEAP_SIZE;
  static_size = STATIC_SIZE;
  generation_size = GENERATION_SIZE;
  arg_stack_size = ARG_STACK_SIZE;
  ret_stack_size = RET_STACK_SIZE;
  verb_gc = 0;

  for(i = 1; argv[i][0] == '-'; i++)
    switch(argv[i][1]) {
    case 'm':
      total_size = atoi(argv[++i]) * 256;
      break;
    case 'h':
      heap_size = atoi(argv[++i]) * 256;
      break;
    case 'g':
      generation_size = atoi(argv[++i]) * 256;
      break;
    case 's':
      static_size = atoi(argv[++i]) * 256;
      break;
    case 'a':
      arg_stack_size = atoi(argv[++i]) * 256;
      break;
    case 'r':
      ret_stack_size = atoi(argv[++i]) * 256;
      break;
    case 'v':
      verb_gc = 1;
      break;
    default:
      fatal_error("unknown option");
    }

  if (argv[i] == 0) {
    write(2, usage, strlen(usage));
    exit(2);
  }

  if (setjmp(raise_buf) == 0) {
    external_raise = (jmp_buf *) raise_buf;
    
    fd = open(argv[i], O_RDONLY);
    if (fd == -1)
      fatal_unix_error("Cannot open ", argv[i]);
    read_header(fd, &hdr);
    
    reset_roots();
    n = (hdr.code_size + 3) / 4;
    static_size += generation_size * 3 / 2 + arg_stack_size + ret_stack_size + n;
    n = (hdr.data_size + 3) / 4 + 1024;
    heap_size += n;
    
    init_memory(total_size, heap_size, generation_size, static_size);
    init_stacks(arg_stack_size, ret_stack_size);
    init_atoms();
    
    read_code(fd, hdr.code_size);
    read_global_data(fd, hdr.data_size);
    close(fd);
    
    unix_init(argc - i, argv + i);
    init_signals();
    
    interprete(start_code, MLINT(0));
    unix_exit(0);
  } else {
    if (exn_bucket == atom(OUT_OF_MEMORY))
      fatal_error ("Out of memory");
    else
      fatal_error ("Uncaught exception");
  }
}


